import logging
import time
from concurrent.futures import ThreadPoolExecutor, as_completed

from astrometry_net_client.session import Session
from astrometry_net_client.settings import Settings
from astrometry_net_client.uploads import FileUpload

log = logging.getLogger(__name__)

# TODO: document this value.
MAX_WORKERS = 10


class Client:
    """
    Higher level class which makes the interaction with the API easy.

    Parameters
    ----------
    session: :py:class:`astrometry_net_client.session.Session`
        Optional argument to provide the session for the API. If not provided,
        you must either specify the keyword ``api_key`` or ``location``, or
        have the key present in your environment. See
        :py:class:`astrometry_net_client.session.Session`
    settings: :py:class:`astrometry_net_client.settings.Settings`, dict
        An optional settings object, which will be applied to all uploads.
        If not specified, the settings object will be created from given
        keyword arguments which correspond to valid settings. For this see
        :py:class:`astrometry_net_client.settings.Settings`.
    kwargs: arguments
        Used to create a session or settings object, if either is not
        specified. Will extract the relevant arguments relevant to the object
        which is created before passing them to the object constructor.
        See :py:class:`astrometry_net_client.settings.Settings` and
        :py:class:`astrometry_net_client.session.Session`
    """

    def __init__(self, session=None, settings=None, **kwargs):
        # TODO, make session optional?
        if session is None:
            args = {
                k: v
                for k, v in kwargs.items()
                if k in {"api_key", "key_location"}
            }
            self.session = Session(**args)
        else:
            self.session = session

        if settings is None:
            setting_args = {
                k: v for k, v in kwargs.items() if k in Settings._settings
            }
            self.settings = Settings(**setting_args)
        else:
            self.settings = Settings(settings)

        log.info("Logging in")
        self.session.login()

    def upload_files_gen(
        self,
        files_iter,
        filter_func=None,
        filter_args=None,
        workers=MAX_WORKERS,
    ):
        """
        Generator which uploads a number of files concurrently, yielding the
        :py:class:`astrometry_net_client.statusables.Job` & filename when done.

        Note that the ``files_iter`` argument is fully consumed when submitting
        to the executor is done. This means that the results will only be
        yielded once the iterator is consumed.

        Parameters
        ----------
        files_iter: iterable
            Some iterable containing paths to the files which will be uploaded.
            Is fully consumed before any result is yielded.
        workers: int, optional
            A positive integer, controlling the amount of workers to use for
            the processing. Will not exceed the value of
            :py:const:`MAX_WORKERS`.
        filter_func: Callable
            Predicate filter function which takes in the `filename` and
            optionally some argument (`filter_args`).
        filter_args: List
            Arguments which are to be passed to the filter function.

        Yields
        ------
        (:py:class:`astrometry_net_client.statusables.Job`, ``str``):
            A tuple containing the finished
            :py:class:`astrometry_net_client.statusables.Job` and the
            corresponding filename. Yields when the Job is finished.
            NOTE: Order of yielded filenames can (and likely will) be different
            from the given ``files_iter``
        """
        workers = min(MAX_WORKERS, workers)
        with ThreadPoolExecutor(max_workers=workers) as executor:
            log.info("Spawned executor {}".format(executor))

            # submit the files & save which future corresponds to which
            #  filename
            future_to_file = {
                executor.submit(
                    self.filtered_upload_wrapper,
                    filename,
                    filter_func=filter_func,
                    filter_args=filter_args,
                ): filename
                for filename in files_iter
            }

            # iterate over the results once they are completed.
            for future in as_completed(future_to_file):
                result_filename = future_to_file[future]
                try:
                    res_job = future.result()
                except Exception as e:
                    # This exception is thrown inside the computed function.
                    err_msg = "File {} stopped with exception {}"
                    log.error(err_msg.format(result_filename, e))
                else:
                    if res_job is not None:  # ignore if file was filtered out
                        yield res_job, result_filename

    def filtered_upload_wrapper(
        self, filename, filter_func=None, filter_args=None, *args, **kwargs
    ):
        """
        Wrapper around :py:func:`upload_file` which filters the given file
        based on a specified filter function.  Main use for this is a
        computationally heavy filter function, like counting number of sources
        locally, and only uploading if not enough are detected.

        Parameters
        ----------
        filename: str
            File to be uploaded. See :py:func:`upload_file`.
        filter_func: Callable
            Predicate filter function which takes in the `filename` and
            optionally some argument (`filter_args`).
        filter_args: List
            Arguments which are to be passed to the filter function.
        args: other arguments
            Directly passed to :py:func:`upload_file`
        kwargs: keyword arguments
            Directly passed to :py:func:`upload_file`

        Returns
        -------
        Job or None: :py:class:`astrometry_net_client.statusables.Job`, `None`
            Will be the job of the resulting upload (see
            :py:func:`upload_file`), or `None` when `filter_func` evaluated to
            `False`.
        """
        if filter_args is None:
            # allow arguments to be unpackable if it is not specified
            filter_args = []

        if filter_func is not None and not filter_func(filename, *filter_args):
            log.info("Filter function failed, skipping upload")
            return None

        return self.upload_file(filename, *args, **kwargs)

    def upload_file(self, filename, settings=None):
        """
        Uploads file and returns completed job when finished solving.

        Parameters
        ----------
        filename: str
            Location + name of the file to upload.
        settings: :py:class:`astrometry_net_client.settings.Settings`
            An optional settings dict which only applies to this specific
            upload. Will override the default settings with which the
            :py:class:`Client` was constructed.

        Returns
        -------
        Job: :py:class:`astrometry_net_client.statusables.Job`
            The job of the resulting upload. NOTE: It is possible that the job
            did not succeed, therefore check with
            :py:meth:`astrometry_net_client.statusables.Statusable.success` if
            it did.
        """
        upl_settings = self.settings
        if settings is not None:
            upl_settings.update(settings)

        upl = FileUpload(filename, session=self.session, settings=upl_settings)
        start = time.time()
        submission = upl.submit()

        msg = "File {} submitted, waiting for it to finish"
        log.info(msg.format(filename))

        submission.until_done()  # blocks here

        # pretty much guarenteed to have exactly one job
        job = submission.jobs[0]

        job.until_done()  # blocks here
        end = time.time()

        msg = "Processing of file {} finished in {}s"
        log.info(msg.format(filename, end - start))

        return job

    def calibrate_file_wcs(self, filename, settings=None):
        """
        Uploads a file, returning the wcs header if it succeeds.

        Parameters
        ----------
        filename: str
            Location + name of the file to upload.
        settings: :py:class:`astrometry_net_client.settings.Settings`
            An optional settings dict which only applies to this specific
            upload. Will override the default settings with which the
            :py:class:`Client` was constructed.

        Returns
        -------
        :py:class:`astropy.io.fits.Header` or None
            Returns the Header if successful, None otherwise.
        """
        job = self.upload_file(filename, settings=settings)

        if job.success():
            return job.wcs_file()
        return None



import os

from astrometry_net_client.config import login_url, read_api_key
from astrometry_net_client.exceptions import (
    APIKeyError,
    InvalidRequest,
    LoginFailedException,
)
from astrometry_net_client.request import PostRequest


class Session(object):
    """
    Class to hold information on the Astrometry.net API session. Is able to
    read an API key from multiple possible locations, and login to retrieve
    a session key (stored in the id attribute).

    Attributes
    ----------
    logged_in : bool
        Boolean indicating if a :py:func:`login` has happened successfully
        since creation of the class. Does not guarentee that the session
        :py:attr:`key` is (still) valid.
    api_key : str
        The API key used to log in.
    key : str
        The session key (not the :py:attr:`api_key`) which is given by the API
        after a valid :py:func:`login`.

    Examples
    --------
    Normal usage of a session is as follows; you create the Session class, and
    then :py:func:`login`:

    >>> session = Session('the api key')
    >>> session.login()

    There are multiple ways of giving your api key. You can just give it as
    a plain string (not very safe):

    >>> session_1 = Session(api_key='the api key')

    You can enter it in a separate private file and give the path to it. This
    file should only contain the key, nothing else.

    >>> session_2 = Session(key_location='path/to/file/with/key')

    You can specify the key as an environment variable named
    ``ASTROMETRY_API_KEY``.

    >>> session_3 = Session()

    If you do not specify any of the above, an exception will be thrown:

    >>> Session()
    APIKeyError

    Raises
    ------
    :py:exc:`APIKeyError`
        When no API key is specified (see examples)
    """

    url = login_url

    def __init__(self, api_key: str = None, key_location: str = None):
        if api_key is not None:
            self.api_key = api_key.strip()
        elif key_location is not None:
            self.api_key = read_api_key(key_location)
        else:
            self.api_key = os.environ.get("ASTROMETRY_API_KEY")
            if self.api_key is None:
                raise APIKeyError(
                    "No api key found or given. "
                    "Specify an API key using one of: "
                    "1. A direct argument, "
                    "2. A location of a file containing the key, "
                    "3. An environment variable named ASTROMETRY_API_KEY."
                )
        self.logged_in = False

    def login(self, force: bool = False):
        """
        Method used to log-in or start a session with the Astrometry.net API.

        Only sends a request if it is absolutely needed (or forced to).  If
        :py:attr:`logged_in` is True, it assumes the session is still valid
        (there is no way to check if the session is still valid).

        After a successful login, the session key is stored in the
        :py:attr:`key` attribute and :py:attr:`logged_in` is set to `True`.

        Parameters
        ----------
        force: bool
            Forces the login by ignoring the :py:attr:`logged_in` boolean.

        Raises
        ------
        LoginFailedException
            If the login response does not have the status 'success'.
        """
        if not force and self.logged_in:
            return

        r = PostRequest(self.url, data={"apikey": self.api_key})
        try:
            response = r.make()
        except InvalidRequest:
            raise LoginFailedException("The api key given is not valid")

        self.key = response["session"]
        self.logged_in = True



from astrometry_net_client.client import Client
from astrometry_net_client.session import Session
from astrometry_net_client.settings import Settings
from astrometry_net_client.statusables import Job, Submission
from astrometry_net_client.uploads import FileUpload

__all__ = ["Settings", "Session", "Job", "Submission", "Client", "FileUpload"]



import json
import logging

import requests
from astropy.io import fits

from astrometry_net_client.exceptions import (
    InvalidRequest,
    InvalidSessionError,
    NoSessionError,
    UnkownContentError,
)

log = logging.getLogger(__name__)


class Request(object):
    """
    Class to make requests to the Astrometry.net API. Intended use mainly for
    extending / internal class usage.

    Requests are made using the requests library, and is capable of making GET
    and POST requests (controlled by the method argument in the constructor).

    Typical usage (makes a GET request):
     >>> r = Request('some-url')
     >>> response = r.make()

    The constructor prepares the request, but does not yet make it. To send /
    make the request use the method: `r.make()`.

    For a POST request:
     >>> r = Request('some-url', method='post', data=data, settings=settings)
     >>> response = r.make()

    where data and settings are both combined and wrapped into the
    'request-json' form, but are split to allow for some general settings.

    An alternative for not using the "method='post'" is to use the PostRequest
    class

    It is valid to omit specifying a url, if the subclass already has its own
    url attribute.

    The internal _make_request() method can be used to extend the functionality
    of requests (e.g. AuthorizedRequest: by making sure the user is logged in).

    Additional arguments specied in the constructor (which are not directly
    used in this class or any subclass) will be stored and passed to the
    request call.

    The original response (as returned from the requests module call) is stored
    in the `original_response` attribute.
    """

    _allowed_methods = {"post", "get"}
    _raw_content_types = {"application/fits", "image/jpeg", "image/png"}

    def __init__(
        self, url=None, method="get", data=None, settings=None, **kwargs
    ):
        self.data = {} if data is None else data.copy()
        self.settings = {} if settings is None else settings.copy()
        self.arguments = kwargs
        if url is not None:
            self.url = url

        # We have to get the attribute like this, otherwise we cannot
        # monkeypatch (e.g. replace requests.get) in the test cases
        if method in self._allowed_methods:
            self.method = getattr(requests, method)
        else:
            err_msg = "Method argument must be one of {}"
            raise ValueError(err_msg.format(self._allowed_methods))

        self.response = None

    def _make_request(self):
        payload = {"request-json": json.dumps({**self.data, **self.settings})}

        log.debug("Sending {!r} with payload {}".format(self, payload))
        response = self.method(self.url, data=payload, **self.arguments)
        log.debug("Retrieved response: {}".format(response.text))

        self.original_response = response

        if response.status_code != 200:
            response.raise_for_status()

        content_type = response.headers["Content-Type"]

        # case where the response is JSON as text
        if content_type.startswith("text/plain"):
            log.debug("Text response detected")
            response = response.json()
            self.response = response

            # TODO add complete response checking
            if response.get("status", "") == "error":
                err_msg = response["errormessage"]
                if err_msg == 'no "session" in JSON.':
                    # No session argument provided
                    raise NoSessionError(err_msg)
                if err_msg.startswith("no session with key"):
                    # Invalid / Expired session key provided
                    raise InvalidSessionError(err_msg)

                # fallback exception for a general error
                raise InvalidRequest(err_msg)

        # case where a file is returned, we want to return the raw bytes
        elif content_type in self._raw_content_types:
            log.debug("Raw file response detected")
            self.response = response.content
        else:
            log.error("Unknown response content {}".format(content_type))
            msg = "Request produced a response with unknown content type {}"
            raise UnkownContentError(msg.format(content_type))

        return self.response

    def make(self):
        """
        Send the actual request to the API.
        returns the response

        This method is mainly for convenience, as to avoid the user unfriendly
        _make_request() method.

        Returns
        -------
        response
            Type depends on the request which is being send. Can be a ``dic``
            but also a binary file (e.g. when expecting a filts file).
        """
        return self._make_request()

    def __repr__(self):
        msg = (
            "Request({self.url}, method={self.method}, data={self.data},"
            " settings={self.settings}, args={self.arguments})"
        )
        return msg.format(self=self)


class PostRequest(Request):
    """
    Makes a POST request instead of the default GET request. It can be used
    as a mixin class, alongside other

    Essentially replaces:
     >>> r = Request(url, method='post', arguments...)

    With
     >>> r = PostRequest(url, arguments...)

    For further usage see the Request class
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, method="post", **kwargs)


class AuthorizedRequest(Request):
    """
    Wraps the normal Request around an authentication layout, ensuring the
    user is logged in and the session key is send alongside the request.

    The separate login request (if needed) is only send just before the
    original request is made, (e.g. when calling make / _make_request).

    Attributes
    ----------
    session: :py:class:`Session`
    """

    def __init__(self, session, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.session = session

    def _make_request(self):
        self.session.login()
        self.data["session"] = self.session.key
        try:
            return super()._make_request()
        except InvalidSessionError:
            log.info("Session expired, loggin in again")
            self.session.login(force=True)
            # update the session key for the request as well
            self.data["session"] = self.session.key

        return super()._make_request()


def file_request(url):
    """
    Utility function which makes a request to ``url`` and gets a file in
    response.

    Parameters
    ----------
    url: str
        URL to send the request to.

    Returns
    -------
        the binary contents of this file.
    """
    r = Request(url)
    binary_file = r.make()
    return binary_file


def fits_file_request(url):
    """
    Make request to ``url`` and return a FITS file.

    Makes a request to ``url``  and read the binary_fits response into
    an astropy fits file (``astropy.io.fits.HDUList``)

    Parameters
    ----------
    url: str
        URL to send the request to.

    Returns
    -------
        an astropy fits file (``astropy.io.fits.HDUList``)
    """
    binary_fits = file_request(url)
    hdul = fits.HDUList(file=binary_fits)
    return hdul



class APIKeyError(Exception):
    pass


# Request exceptions
class RequestException(Exception):
    pass


class InvalidRequest(RequestException):
    pass


class FailedRequest(RequestException):
    pass


class LoginFailedException(RequestException):
    pass


class StillProcessingException(RequestException):
    pass


class StatusFailedException(RequestException):
    pass


class UnkownContentError(RequestException):
    pass


# Session exceptions
class SessionError(RequestException):
    pass


class InvalidSessionError(SessionError):
    pass


class NoSessionError(SessionError):
    pass


class ExhaustedAttemptsException(Exception):
    pass



import setuptools

with open("README.rst", "r") as f:
    long_description = f.read()

setuptools.setup(
    name="astrometry_net_client",
    version="0.2.4",
    author="Sten Sipma",
    author_email="sten.sipma@ziggo.nl",
    description="A Python interface for the Astrometry.net API.",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/StenSipma/astrometry_net_client",
    packages=["astrometry_net_client"],
    install_requires=["astropy", "requests"],
    keywords="astronomy astrometry client coordinates wcs api",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Scientific/Engineering :: Astronomy",
        "Topic :: Scientific/Engineering :: Image Recognition",
        "Operating System :: POSIX :: Linux",
        "Natural Language :: English",
    ],
    python_requires=">=3.7",
)



import abc

from astrometry_net_client.config import upload_url
from astrometry_net_client.request import AuthorizedRequest, PostRequest
from astrometry_net_client.statusables import Submission


class Submitter(abc.ABC):
    """
    Abstract class intended to use as a mixin class with a Request object (and
    hence assumes a `_make_request` method).
    Provides a new `submit` method, similar to `make`, but creates a Submission
    object out of the resulting response (from the subid entry).

    Example
    -------
    >>> upl = FileUpload(filename, session)
    >>> submission = upl.submit()
    """

    @abc.abstractmethod
    def _make_request(self):
        pass

    def submit(self):
        """
        Submit function, similar to ``make`` but in addition creates and
        returns the resulting submission.

        Returns
        -------
        :py:class:`astrometry_net_client.statusables.Submission`
            The submission which is created by the API. Use this to get the
            status & result(s) of your upload.
        """
        response = self.make()
        return Submission(response["subid"])


class BaseUpload(AuthorizedRequest, PostRequest, Submitter):
    """
    Extends from:

    #. :py:class:`astrometry_net_client.request.AuthorizedRequest`,
    #. :py:class:`astrometry_net_client.request.PostRequest`,
    #. :py:class:`Submitter`

    Base class for uploads to be made with the Astrometry.net API. Combines
    some abstact classes useful for an upload request (e.g. a submit method,
    default post request and authorization)

    Meant as an abstact/base class, not to be called directly.
    """

    pass


class FileUpload(BaseUpload):
    """Request for submitting a file

    Extends from :py:class:`BaseUpload`

    Class for uploading a file to Astrometry.net `upload file`_ endpoint

    .. _upload file: http://astrometry.net/doc/net/api.html#submitting-a-file

    Example
    -------
    >>> s = Session(api_key='XXXXX')
    >>> upl = FileUpload('some/file', session=s)
    >>> submission = upl.submit()

    You can then use the resulting submission to inspect the result of your
    upload.

    Parameters
    ----------
    filename : str
        The path to the file which you want to upload.
    session : :py:class:`astrometry_net_client.session.Session`
        The login session (required by the
        :py:class:`astrometry_net_client.request.AuthorizedRequest`
        class)

    Returns
    -------
    :py:class:`astrometry_net_client.statusables.Submission`
        The submission object, which is created when you upload a file.
    """

    url = upload_url

    def __init__(self, filename, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # TODO check if file exists?
        # TODO check if file has the correct type
        self.filename = filename

    def _make_request(self):
        with open(self.filename, "rb") as f:
            self.arguments["files"] = {"file": f}
            response = super()._make_request()
        return response

    def __repr__(self):
        msg = "FileUpload(filename={}, session={})"
        return msg.format(self.filename, self.session)


class URLUpload(BaseUpload):
    """
    Extends from :py:class:`BaseUpload`

    Class for making a url upload to Astrometry.net
    http://astrometry.net/doc/net/api.html#submitting-a-url

    Work in progress
    """

    # TODO implement


class SourcesUpload(BaseUpload):
    """
    Extends from :py:class:`BaseUpload`

    Class for uploading a list of sources. (inspired by astroquery
    implementation)
    """

    # TODO: Decide if this is necessary



import abc
import logging
import time
from functools import wraps

from astropy.io import fits

from astrometry_net_client.config import BASE_URL
from astrometry_net_client.exceptions import (
    StatusFailedException,
    StillProcessingException,
)
from astrometry_net_client.request import (
    Request,
    file_request,
    fits_file_request,
)

log = logging.getLogger(__name__)


def cache_response(func):
    """
    Wrapper around a function to cache its result in an attribute of the
    object. Name of the attribute is: _<funcname>_result

    Parameters
    ----------
    func
        function to be wrapped

    Returns
    -------
    wrapped function
    """
    func_name = func.__name__
    result_attr = "_{}_result".format(func_name)

    @wraps(func)
    def wrapper(self, *args, force=False, **kwargs):
        if not force and hasattr(self, result_attr):
            log_msg = "Result {} already cached. Reusing..."
            log.debug(log_msg.format(result_attr))
            return getattr(self, result_attr)
        result = func(self, *args, **kwargs)
        setattr(self, result_attr, result)
        return result

    return wrapper


def ensure_status_success(func):
    """
    Decorator for a method to enforce it only being called when the
    statusable is successful (and therefore also finished).

    Parameters
    ----------
    func
        function to be wrapped

    Returns
    -------
    wrapped function

    Raises
    ------
    StatusFailedException
        When the :py:meth:`success` evaluates to ``False`` but
        :py:meth:`done` to ``True``.
    StillProcessingException
        When :py:meth:`done` is ``False``
    """

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        self.status()
        if not self.done():
            raise StillProcessingException()
        if not self.success():
            raise StatusFailedException()
        result = func(self, *args, **kwargs)
        return result

    return wrapper


def ensure_status(func):
    """
    Decorator for a method to enforce it only being called when the
    statusable is finished.

    Parameters
    ----------
    func
        function to be wrapped

    Returns
    -------
    wrapped function

    Raises
    ------
    StillProcessingException
        when :py:meth:`done` is ``False``
    """

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        self.status()
        if not self.done():
            raise StillProcessingException()
        result = func(self, *args, **kwargs)
        return result

    return wrapper


class Statusable(abc.ABC):
    @abc.abstractmethod
    def _make_status_request(self):
        pass

    @abc.abstractmethod
    def _is_final_status(self):
        pass

    @abc.abstractmethod
    def _status_success(self):
        pass

    def status(self, force=False):
        """
        Method which queries the status if it is needed.

        ``status`` is a method which queries the status of the statusable
        if the current / last retrieved status is not :py:meth:`done` (or no
        status is known). If an earlier queried status signals that the
        statusable is finished, no further request is made.

        Parameters
        ----------
        force: bool
            Can be set to true to make a status request regardless of the
            result of :py:meth:`done`.

        Returns
        -------
            dict: Dictionary with the status response (content). Same as
                attribute :py:attr:`stat_response`.
        """
        if force or not self.done():
            log_msg = "Statusable {} not done, querying status..."
            log.debug(log_msg.format(self.__class__.__name__))
            self.stat_response = self._make_status_request()

        return self.stat_response

    def until_done(self, start=4, end=300, timeout=None):
        """
        Blocking method which waits for the Statusable to be finished.

        This method will keep querying the :py:func:`status` until the
        statusable is :py:func:`done`. Will sleep in increasing intervals
        between each status query, to avoid overloading the API server
        (and give room for possible threading). The sleeping behaviour is
        determined by ``start`` and ``end``.

        It is possible to specify a timeout, in seconds, after which a
        :py:exc:`TimeoutError` is raised.

        Parameters
        ----------
        start: int
            Determines the intitial sleep time, which is doubled after each
            query. Defaults to 4s.
        end: int or None
            If specified, gives a maximal value for the sleep time. Otherwise
            the sleep time will be doubled forever.
        timeout: int or None
            If specified will raise a :py:exc:`TimeoutError` when the
            method has been running for the given time.

        Returns
        -------
        dict:
            Dictionary with the status response (content). Same as attribute
            :py:attr:`stat_response`.

        Raises
        ------
        TimeoutError
            When parameter `timeout` is set and the waited time exceeds this
            value.

        Examples
        --------
        Will wait forever and doubles the sleep time until it reaches 300.

        >>> stat.until_done()

        Will wait forever, and doubles the sleep time until it is equal to 60s

        >>> stat.until_done(end=60)

        Will wait forever and doubles the sleep time indefinitely

        >>> stat.until_done(end=None)

        Waits until 120s have passed (default sleep behaviour)

        >>> stat.until_done(timeout=120s)
        TimeoutError
        """
        now = time.time  # alias the function for readablility
        start_time = now()

        sleep_time = start
        log_msg = (
            "Starting the until done loop with: sleep_time = {}, timeout = {}"
        )
        log.debug(log_msg.format(sleep_time, timeout))
        while timeout is None or now() - start_time < timeout:
            response = self.status()
            log.debug("Current status response: {}".format(response))

            if self.done():
                break

            log.debug("Not done, sleeping for {}s".format(sleep_time))
            time.sleep(sleep_time)
            sleep_time *= 2
            sleep_time = end if end and end < sleep_time else sleep_time
        else:
            raise TimeoutError()

        return self.stat_response

    def success(self):
        """
        Evaluates if the statusable was successful.

        Will always be ``False`` if :py:meth:`done` is ``False``.

        Returns
        -------
        bool
        """
        # TODO maybe return None when you cannot yet know ?
        return self._is_final_status() and self._status_success()

    def done(self):
        """
        Evaluates if the last :py:meth:`status` result is the final result.

        Returns
        -------
        bool:
            ``True``, when the last response from `status` is the final status
            (e.g. it does not change anymore). ``False`` otherwise
        """
        return self._is_final_status()


class Submission(Statusable):
    """
    Represents a single submission from Astrometry.net.  A submission is the
    result of any of the :py:class:`astrometry_net_client.UploadFile` or
    :py:class:`astrometry_net_client.UploadURL` uploaders.

    When :py:func:`status` is called and jobs are available, the status of the
    jobs will also be queried. To see the response of these jobs, see their
    ``response`` attribute.

    Note that some of the attributes listed below are only available when
    :py:func:`status` is queried.

    Attributes
    ----------
    id: int
        Identifier for the submission. Used in the request URL.
    user: int
        Identifier for the user which made the Submission.
    images: list(int)
        List of references to the uploaded images.
    job_calibrations: list(int)
        Identifier to the references used to solve the images.
    jobs: list(Jobs)
        The list of jobs which the Submission spawned. Will typically be one
        Job per uploaded image. When no job was spawned yet, the list will
        be: ``[None]``.
    processing_started: str
        If available, a string containing the date when the Submission process
        started.
    processing_finished: str
        If available, a string containing the date when the Submission process
        finished.

    See Also
    --------
    Statusable : For available function on getting the status & waiting
    """

    url = BASE_URL + "/submissions/{submission.id}"

    def __init__(self, submission_id):
        self.id = submission_id
        self.url = self.url.format(submission=self)

    @ensure_status
    def __iter__(self):
        """
        Allowes to easily iterate over the jobs of a submission by doing:
        >>> for job in submission:
        ...    process_job()

        Returns
        -------
        Iterator over the :py:attr:`jobs`.
        """
        return iter(self.jobs)

    def _make_status_request(self):
        r = Request(self.url)
        response = r.make()
        self.response = response

        # TODO verify which of these entries are valid in the response
        self.processing_started = response["processing_started"]
        self.processing_finished = response["processing_finished"]
        self.user_images = response["user_images"]
        self.user = response["user"]
        self.images = response["images"]
        self.job_calibrations = response["job_calibrations"]

        self.jobs = [
            Job(job_id) for job_id in response["jobs"] if job_id is not None
        ]

        for job in self.jobs:
            job.status()

        return response

    def _is_final_status(self):
        """
        We define 'final' here as processing has finished and the jobs have
        started, assuming there is always at least one job.

        Returns
        -------
        bool
        """
        try:
            proc_finished = self.processing_finished is not None
        except AttributeError:
            return False

        jobs_created = hasattr(self, "jobs") and len(self.jobs) > 0
        return proc_finished and jobs_created

    def _status_success(self):
        if self.done():
            return all(job.success() for job in self)
        return False

    def __repr__(self):
        return "Submission({self.id})".format(self=self)

    def __str__(self):
        msg = "Submission(id={}, final={}, success={})"
        return msg.format(self.id, self.done(), self.success())

    def __hash__(self):
        return hash(self.id)


class Job(Statusable):
    """
    Represents a single job from Astrometry.net.
    It can either still be running, or be finished.
    When finished, results can be queried using the appropriate methods.

    Attributes
    ----------
    id: int
        Identifier for the Astrometry.net Job.
    resp_status: str
        The status of the Job (as it was last queried). Can be: ``"success"``,
        ``"failure"`` or ``"solving"``.
    objects_in_field : list(str)
        Detected objects in the field. Result of the :py:meth:`info` query.
    machine_tags : list(str)
        Tags made by the API (machine). Result of the :py:meth:`info` query.
    tags : list(str)
        All tags of the Job. Can be made by users or by the API. Result of
        the :py:meth:`info` query.
    original_filename : str
        Filename of the uploaded file. Will not include the path. Result of
        the :py:meth:`info` query.
    calibration : dict
        Dictionary containing information about the solved image, example:

        >>> job.calibration
        {"parity": 1.0, "orientation": 105.74942079091929,
                "pixscale": 1.0906710701159739, "radius": 0.8106715896625917,
                "ra": 169.96633791366915, "dec": 13.221011585315143}

    See Also
    --------
    Statusable : For available functions on getting the status & waiting
    """

    url = BASE_URL + "/jobs/{job.id}"

    info_suffix = "/info"

    # does not include the .../api/ path ... :(
    wcs_file_url = "http://nova.astrometry.net/wcs_file/{job.id}"
    fits_file_url = "http://nova.astrometry.net/new_fits_file/{job.id}"
    rdls_file_url = "http://nova.astrometry.net/rdls_file/{job.id}"
    axy_file_url = "http://nova.astrometry.net/axy_file/{job.id}"
    corr_file_url = "http://nova.astrometry.net/corr_file/{job.id}"
    annotated_display_url = (
        "http://nova.astrometry.net/annotated_display/{job.id}"
    )
    red_green_image_display_url = (
        "http://nova.astrometry.net/red_green_image_display/{job.id}"
    )
    extraction_image_display_url = (
        "http://nova.astrometry.net/extraction_image_display/{job.id}"
    )

    def __init__(self, job_id):
        self.id = job_id
        self.url = self.url.format(job=self)

    def _make_status_request(self):
        r = Request(self.url)
        response = r.make()
        self.resp_status = response["status"]
        return response

    def _is_final_status(self):
        try:
            return self.resp_status in {"success", "failure"}
        except AttributeError:
            return False

    def _status_success(self):
        return self.resp_status == "success"

    @ensure_status
    @cache_response
    def info(self):
        r = Request(self.url + self.info_suffix)
        response = r.make()

        self.objects_in_field = response["objects_in_field"]
        self.machine_tags = response["machine_tags"]
        self.tags = response["tags"]
        self.original_filename = response["original_filename"]
        # TODO make a calibration class ?
        if self.success():
            self.calibration = response["calibration"]

        return response

    @ensure_status_success
    @cache_response
    def wcs_file(self):
        """
        Get the resulting wcs file as an astropy.io.fits.Header.

        Returns
        -------
        astropy.io.fits.Header
        """
        r = Request(self.wcs_file_url.format(job=self))
        binary_wcs = r.make()
        header = fits.Header.fromstring(binary_wcs)
        return header

    @ensure_status_success
    @cache_response
    def new_fits_file(self):
        """
        Get the new fits file (original file + new header).

        Returns
        -------
        astropy.io.fits.HDUList
        """
        return fits_file_request(self.fits_file_url.format(job=self))

    @ensure_status_success
    @cache_response
    def rdls_file(self):
        return fits_file_request(self.rdls_file_url.format(job=self))

    @ensure_status_success
    @cache_response
    def axy_file(self):
        return fits_file_request(self.axy_file_url.format(job=self))

    @ensure_status_success
    @cache_response
    def corr_file(self):
        return fits_file_request(self.corr_file_url.format(job=self))

    @ensure_status_success
    @cache_response
    def annotated_display(self):
        """
        JPEG image as a binary string

        Returns
        -------
        bytes
        """
        return file_request(self.annotated_display_url.format(job=self))

    @ensure_status_success
    @cache_response
    def red_green_image_display(self):
        """
        PNG image as a binary string

        Returns
        -------
        bytes
        """
        return file_request(self.red_green_image_display_url.format(job=self))

    @ensure_status_success
    @cache_response
    def extraction_image_display(self):
        """
        PNG image as a binary string

        Returns
        -------
        bytes
        """
        return file_request(self.extraction_image_display_url.format(job=self))

    def __repr__(self):
        return "Job({self.id})".format(self=self)

    def __str__(self):
        msg = "Job(id={}, final={}, success={})"
        return msg.format(self.id, self.done(), self.success())

    def __hash__(self):
        return hash(self.id)
